@@ -27,6 +27,9 @@ import six
 from six.moves.urllib.parse import quote
 
 from AccessControl.class_init import InitializeClass
+from AccessControl.Permissions import change_proxy_roles
+from AccessControl.Permissions import change_python_scripts
+from AccessControl.Permissions import view_management_screens
 from AccessControl.requestmethod import requestmethod
 from AccessControl.SecurityInfo import ClassSecurityInfo
 from AccessControl.SecurityManagement import getSecurityManager
@@ -37,12 +40,15 @@ from App.Common import package_home
 from App.Dialogs import MessageDialog
 from App.special_dtml import DTMLFile
 from OFS.Cache import Cacheable
+from OFS.History import Historical
+from OFS.History import html_diff
 from OFS.SimpleItem import SimpleItem
 from RestrictedPython import compile_restricted_function
 from Shared.DC.Scripts.Script import BindingsUI
 from Shared.DC.Scripts.Script import Script
 from Shared.DC.Scripts.Script import defaultBindings
 from zExceptions import Forbidden
+from ZPublisher.HTTPRequest import default_encoding
 
 
 try:
@@ -74,7 +80,8 @@ _default_file = os.path.join(package_home(globals()), 'www', 'default_content')
 _marker = []  # Create a new marker object
 
 
-def manage_addPythonScript(self, id, title='', REQUEST=None, submit=None):
+def manage_addPythonScript(self, id, title='', file=None, REQUEST=None,
+                           submit=None):
     """Add a Python script to a folder.
     """
     id = str(id)
@@ -82,13 +89,16 @@ def manage_addPythonScript(self, id, title='', REQUEST=None, submit=None):
     pyscript = self._getOb(id)
     if title:
         pyscript.ZPythonScript_setTitle(title)
+
+    file = file or (REQUEST and REQUEST.form.get('file'))
+    if hasattr(file, 'read'):
+        file = file.read()
+    if not file:
+        with open(_default_file) as fp:
+            file = fp.read()
+    pyscript.write(file)
+
     if REQUEST is not None:
-        file = REQUEST.form.get('file', '')
-        if not isinstance(file, str):
-            file = file.read()
-        if not file:
-            file = open(_default_file).read()
-        pyscript.write(file)
         try:
             u = self.DestinationURL()
         except Exception:
@@ -99,7 +109,7 @@ def manage_addPythonScript(self, id, title='', REQUEST=None, submit=None):
     return ''
 
 
-class PythonScript(Script, Cacheable):
+class PythonScript(Script, Historical, Cacheable):
     """Web-callable scripts written in a safe subset of Python.
 
     The function may include standard python code, so long as it does
@@ -119,35 +129,36 @@ class PythonScript(Script, Cacheable):
     ) + BindingsUI.manage_options + (
         {'label': 'Test', 'action': 'ZScriptHTML_tryForm'},
         {'label': 'Proxy', 'action': 'manage_proxyForm'},
-    ) + SimpleItem.manage_options + \
+    ) + Historical.manage_options + SimpleItem.manage_options + \
         Cacheable.manage_options
 
     def __init__(self, id):
-        self.id = id
         self.ZBindings_edit(defaultBindings)
+        bind_names = self.getBindingAssignments().getAssignedNamesInOrder()
+        if id in bind_names:
+            raise ValueError(
+                'Following names are not allowed as identifiers, as they'
+                'have a special meaning for PythonScript: '
+                '%s.'
+                'Please choose another name.' % ', '.join(bind_names),
+                )
+        self.id = id
         self._makeFunction()
 
     security = ClassSecurityInfo()
 
     security.declareObjectProtected('View')
-    security.declareProtected('View', '__call__')  # NOQA: flake8: D001
+    security.declareProtected('View', '__call__')  # noqa: D001
 
-    security.declareProtected(  # NOQA: flake8: D001
-        'View management screens',
-        'ZPythonScriptHTML_editForm', 'manage_main', 'read',
-        'ZScriptHTML_tryForm', 'PrincipiaSearchSource',
-        'document_src', 'params', 'body', 'get_filepath')
+    security.declareProtected(  # noqa: D001
+        view_management_screens,
+        'ZPythonScriptHTML_editForm', 'manage_main', 'ZScriptHTML_tryForm')
 
     ZPythonScriptHTML_editForm = DTMLFile('www/pyScriptEdit', globals())
     manage = manage_main = ZPythonScriptHTML_editForm
     ZPythonScriptHTML_editForm._setName('ZPythonScriptHTML_editForm')
 
-    security.declareProtected(  # NOQA: flake8: D001
-        'Change Python Scripts',
-        'ZPythonScriptHTML_editAction',
-        'ZPythonScript_setTitle', 'ZPythonScript_edit',
-        'ZPythonScriptHTML_upload')
-
+    @security.protected(change_python_scripts)
     def ZPythonScriptHTML_editAction(self, REQUEST, title, params, body):
         """Change the script's main parameters."""
         self.ZPythonScript_setTitle(title)
@@ -156,12 +167,14 @@ class PythonScript(Script, Cacheable):
         return self.ZPythonScriptHTML_editForm(self, REQUEST,
                                                manage_tabs_message=message)
 
+    @security.protected(change_python_scripts)
     def ZPythonScript_setTitle(self, title):
         title = str(title)
         if self.title != title:
             self.title = title
             self.ZCacheable_invalidate()
 
+    @security.protected(change_python_scripts)
     def ZPythonScript_edit(self, params, body):
         self._validateProxy()
         if self.wl_isLocked():
@@ -173,6 +186,7 @@ class PythonScript(Script, Cacheable):
             self._params = str(params)
             self.write(body)
 
+    @security.protected(change_python_scripts)
     def ZPythonScriptHTML_upload(self, REQUEST, file=''):
         """Replace the body of the script with the text in file."""
         if self.wl_isLocked():
@@ -201,6 +215,13 @@ class PythonScript(Script, Cacheable):
                 param_names.append(name.split('=', 1)[0].strip())
         return param_names
 
+    @security.protected(view_management_screens)
+    def manage_historyCompare(self, rev1, rev2, REQUEST,
+                              historyComparisonResults=''):
+        return PythonScript.inheritedAttribute('manage_historyCompare')(
+            self, rev1, rev2, REQUEST,
+            historyComparisonResults=html_diff(rev1.read(), rev2.read()))
+
     def __setstate__(self, state):
         Script.__setstate__(self, state)
         if getattr(self, 'Python_magic', None) != Python_magic or \
@@ -349,6 +370,7 @@ class PythonScript(Script, Cacheable):
         # shut up deprecation warnings
         pass
 
+    @security.protected(view_management_screens)
     def get_filepath(self):
         return self.meta_type + ':' + '/'.join(self.getPhysicalPath())
 
@@ -368,12 +390,12 @@ class PythonScript(Script, Cacheable):
             'because you do not have proxy roles.\n<!--%s, %s-->' % (
                 self.id, user, roles))
 
-    security.declareProtected(  # NOQA: flake8: D001
-        'Change proxy roles',
-        'manage_proxyForm', 'manage_proxy')
+    security.declareProtected(change_proxy_roles,  # NOQA: D001
+                              'manage_proxyForm')
 
     manage_proxyForm = DTMLFile('www/pyScriptProxy', globals())
 
+    @security.protected(change_proxy_roles)
     @requestmethod('POST')
     def manage_proxy(self, roles=(), REQUEST=None):
         """Change Proxy Roles"""
@@ -387,20 +409,24 @@ class PythonScript(Script, Cacheable):
                 message='Your changes have been saved',
                 action='manage_main')
 
-    security.declareProtected(  # NOQA: flake8: D001
-        'Change Python Scripts',
-        'PUT', 'manage_FTPput', 'write')
+    security.declareProtected(  # NOQA: D001
+        change_python_scripts,
+        'manage_FTPput', 'manage_historyCopy',
+        'manage_beforeHistoryCopy', 'manage_afterHistoryCopy')
 
+    @security.protected(change_python_scripts)
     def PUT(self, REQUEST, RESPONSE):
         """ Handle HTTP PUT requests """
         self.dav__init(REQUEST, RESPONSE)
         self.dav__simpleifhandler(REQUEST, RESPONSE, refresh=1)
-        self.write(REQUEST.get('BODY', ''))
+        new_body = REQUEST.get('BODY', '')
+        self.write(new_body)
         RESPONSE.setStatus(204)
         return RESPONSE
 
     manage_FTPput = PUT
 
+    @security.protected(change_python_scripts)
     def write(self, text):
         """ Change the Script by parsing a read()-style source text. """
         self._validateProxy()
@@ -408,6 +434,11 @@ class PythonScript(Script, Cacheable):
         bindmap = self.getBindingAssignments().getAssignedNames()
         bup = 0
 
+        if six.PY3 and isinstance(text, bytes):
+            text = text.decode(default_encoding)
+        elif six.PY2 and not isinstance(text, bytes):
+            text = text.encode(default_encoding)
+
         st = 0
         try:
             while 1:
@@ -459,11 +490,13 @@ class PythonScript(Script, Cacheable):
             LOG.error('write failed', exc_info=sys.exc_info())
             raise
 
-    def manage_FTPget(self):
-        """Get source for FTP download"""
+    def manage_DAVget(self):
+        """Get source for WebDAV"""
         self.REQUEST.RESPONSE.setHeader('Content-Type', 'text/plain')
         return self.read()
 
+    manage_FTPget = manage_DAVget
+
     def _metadata_map(self):
         m = {
             'title': self.title,
@@ -474,6 +507,7 @@ class PythonScript(Script, Cacheable):
             m['bind ' + k] = bindmap.get(v, '')
         return m
 
+    @security.protected(view_management_screens)
     def read(self):
         """ Generate a text representation of the Script source.
 
@@ -504,9 +538,11 @@ class PythonScript(Script, Cacheable):
         hlines.append('')
         return ('\n' + prefix).join(hlines) + '\n' + self._body
 
+    @security.protected(view_management_screens)
     def params(self):
         return self._params
 
+    @security.protected(view_management_screens)
     def body(self):
         return self._body
 
@@ -515,10 +551,12 @@ class PythonScript(Script, Cacheable):
 
     getSize = get_size
 
+    @security.protected(view_management_screens)
     def PrincipiaSearchSource(self):
         """Support for searching - the document's contents are searched."""
         return '%s\n%s' % (self._params, self._body)
 
+    @security.protected(view_management_screens)
     def document_src(self, REQUEST=None, RESPONSE=None):
         """Return unprocessed document source."""
 
